#!/usr/bin/env python
# License: GPLv3 Copyright: 2024, Kovid Goyal <kovid at kovidgoyal.net>

from . import BaseTest


class TestCommandPalette(BaseTest):

    def test_collect_keys_data(self):
        from kittens.command_palette.main import collect_keys_data
        from kitty.actions import groups
        opts = self.set_options()
        data = collect_keys_data(opts)
        self.assertIn('modes', data)
        self.assertIn('mouse', data)
        self.assertIn('', data['modes'], 'Default keyboard mode should be present')
        default_mode = data['modes']['']
        # Should have at least some categories
        self.assertTrue(len(default_mode) > 0, 'Should have at least one category')
        # All category names should be from the known groups or special palette sections
        known_titles = set(groups.values()) | {'Action aliases', 'Kitten aliases', 'Combined actions'}
        for cat_name in default_mode:
            self.assertIn(cat_name, known_titles, f'Unknown category: {cat_name}')
        # Each category should have bindings with required fields
        for cat_name, bindings in default_mode.items():
            self.assertIsInstance(bindings, list)
            for b in bindings:
                self.assertIn('key', b)
                self.assertIn('action', b)
                self.assertIn('action_display', b)
                self.assertIn('definition', b)
                self.assertIn('help', b)
                self.assertIn('long_help', b)
                self.assertIsInstance(b['key'], str)
                self.assertIsInstance(b['action'], str)
                # key may be empty for unmapped actions; action must always be non-empty
                self.assertTrue(len(b['action']) > 0)
        # Mouse mappings
        self.assertIsInstance(data['mouse'], list)
        for b in data['mouse']:
            self.assertIn('key', b)
            self.assertIn('action', b)
            self.assertIn('action_display', b)

    def test_collect_keys_categories_ordered(self):
        from kittens.command_palette.main import collect_keys_data
        from kitty.actions import groups
        opts = self.set_options()
        data = collect_keys_data(opts)
        default_mode = data['modes']['']
        cat_names = list(default_mode.keys())
        group_titles = list(groups.values())
        # Categories should appear in the same order as defined in groups
        indices = []
        for cat in cat_names:
            if cat in group_titles:
                indices.append(group_titles.index(cat))
        self.ae(indices, sorted(indices), 'Categories should be ordered according to groups dict')

    def test_collect_keys_bindings_sorted(self):
        from kittens.command_palette.main import collect_keys_data
        opts = self.set_options()
        data = collect_keys_data(opts)
        # Within each category, mapped entries (non-empty key) come first sorted by key,
        # then unmapped entries (empty key) sorted by action name.
        for cat_name, bindings in data['modes'][''].items():
            seen_unmapped = False
            for b in bindings:
                if b['key'] == '':
                    seen_unmapped = True
                elif seen_unmapped:
                    self.fail(
                        f'In category {cat_name!r}, mapped binding {b!r} follows an unmapped one'
                    )

    def test_collect_keys_has_help_text(self):
        from kittens.command_palette.main import collect_keys_data
        opts = self.set_options()
        data = collect_keys_data(opts)
        # At least some bindings should have help text
        has_help = False
        for cat_name, bindings in data['modes'][''].items():
            for b in bindings:
                if b['help']:
                    has_help = True
                    break
            if has_help:
                break
        self.assertTrue(has_help, 'At least some bindings should have help text')

    def test_ordering_arrays_present(self):
        from kittens.command_palette.main import collect_keys_data
        opts = self.set_options()
        data = collect_keys_data(opts)
        # mode_order should list all modes
        self.assertIn('mode_order', data)
        self.assertIsInstance(data['mode_order'], list)
        self.ae(set(data['mode_order']), set(data['modes'].keys()))
        # category_order should list categories for each mode
        self.assertIn('category_order', data)
        self.assertIsInstance(data['category_order'], dict)
        for mode_name in data['modes']:
            self.assertIn(mode_name, data['category_order'])
            self.ae(
                set(data['category_order'][mode_name]),
                set(data['modes'][mode_name].keys()),
                f'category_order for mode {mode_name!r} should match modes keys',
            )

    def test_always_includes_unmapped_actions(self):
        from kittens.command_palette.main import collect_keys_data
        opts = self.set_options()
        data = collect_keys_data(opts)
        # Unmapped actions (empty key) are always included
        found_unmapped = False
        for cats in data['modes'].values():
            for bindings in cats.values():
                for b in bindings:
                    if b['key'] == '':
                        found_unmapped = True
                        # Unmapped actions must still have action and definition
                        self.assertTrue(len(b['action']) > 0)
                        self.assertTrue(len(b['definition']) > 0)
                        break
        self.assertTrue(found_unmapped, 'Expected at least one unmapped action to always be present')

    def test_alias_resolution(self):
        from kittens.command_palette.main import collect_keys_data
        from kitty.options.utils import ActionAlias, AliasMap, parse_map
        opts = self.set_options()
        # Set up action aliases: launch_tab (bound) and launch_bg (unbound)
        alias_map = AliasMap()
        alias_map.append('launch_tab', ActionAlias('launch_tab', 'launch --type=tab --cwd=current'))
        alias_map.append('launch_bg', ActionAlias('launch_bg', 'launch --type=background'))
        opts.alias_map = alias_map
        # Add a keybinding that uses the launch_tab alias
        for kd in parse_map('f1 launch_tab vim'):
            kd = kd.resolve_and_copy(opts.kitty_mod)
            default_mode = opts.keyboard_modes['']
            default_mode.keymap.setdefault(kd.trigger, []).append(kd)

        data = collect_keys_data(opts)

        # Aliases should have their own section
        self.assertIn('Action aliases', data['modes'][''],
                       'Aliases should have a dedicated section')
        alias_section = data['modes']['']['Action aliases']

        # Bound alias should appear in the alias section with its key
        bound = [b for b in alias_section if b.get('alias') == 'launch_tab']
        self.ae(len(bound), 1, 'Bound alias should appear exactly once')
        self.ae(bound[0]['action'], 'launch')
        self.assertTrue(bound[0]['key'] != '', 'Bound alias should have its key')

        # Unbound alias should also appear in the alias section
        unbound = [b for b in alias_section if b['action_display'] == 'launch_bg']
        self.ae(len(unbound), 1, 'Unbound alias should appear exactly once')
        self.ae(unbound[0]['action'], 'launch')
        self.ae(unbound[0]['key'], '')
        self.assertTrue(unbound[0]['help'].startswith('Alias for:'))

        # Bound alias should NOT appear in any other category
        for cat_name, bindings in data['modes'][''].items():
            if cat_name == 'Action aliases':
                continue
            for b in bindings:
                self.assertNotEqual(b.get('alias'), 'launch_tab',
                                    f'Alias binding should not appear in {cat_name!r}')

    def test_kitten_alias_section(self):
        from kittens.command_palette.main import collect_keys_data
        from kitty.options.utils import ActionAlias, AliasMap
        opts = self.set_options()
        # Set up a kitten alias: kitten hints -> kitten hints --hints-offset=0
        alias_map = AliasMap()
        alias_map.append('kitten', ActionAlias('hints', 'hints --hints-offset=0', replace_second_arg=True))
        opts.alias_map = alias_map

        data = collect_keys_data(opts)
        self.assertIn('Kitten aliases', data['modes'][''],
                       'Kitten aliases should have a dedicated section')
        kitten_section = data['modes']['']['Kitten aliases']
        found = [b for b in kitten_section if b['action_display'] == 'kitten hints']
        self.ae(len(found), 1, 'Kitten alias should appear exactly once')
        self.ae(found[0]['definition'], 'kitten hints --hints-offset=0')
        self.assertTrue(found[0]['help'].startswith('Alias for:'))

    def test_combine_actions_section(self):
        from kittens.command_palette.main import collect_keys_data
        from kitty.options.utils import parse_map
        opts = self.set_options()
        # Add a combine keybinding
        for kd in parse_map('f2 combine : new_tab : launch vim'):
            kd = kd.resolve_and_copy(opts.kitty_mod)
            default_mode = opts.keyboard_modes['']
            default_mode.keymap.setdefault(kd.trigger, []).append(kd)

        data = collect_keys_data(opts)
        # Combine bindings should have their own section
        self.assertIn('Combined actions', data['modes'][''],
                       'Combine bindings should have a dedicated section')
        combine_section = data['modes']['']['Combined actions']
        found = [b for b in combine_section if b['action'] == 'combine']
        self.assertTrue(len(found) > 0, 'Combine binding should be in the section')
        self.ae(found[0]['key'], 'f2')

    def test_no_duplicate_alias_entries(self):
        from kittens.command_palette.main import collect_keys_data
        from kitty.options.utils import ActionAlias, AliasMap, parse_map
        opts = self.set_options()
        # Set up aliases, some bound to keys and some not
        alias_map = AliasMap()
        alias_map.append('launch_tab', ActionAlias('launch_tab', 'launch --type=tab --cwd=current'))
        alias_map.append('launch_bg', ActionAlias('launch_bg', 'launch --type=background'))
        opts.alias_map = alias_map
        for kd in parse_map('f1 launch_tab vim'):
            kd = kd.resolve_and_copy(opts.kitty_mod)
            default_mode = opts.keyboard_modes['']
            default_mode.keymap.setdefault(kd.trigger, []).append(kd)

        data = collect_keys_data(opts)
        # Collect all entries that have an alias field across all categories
        alias_entries: list[tuple[str, str]] = []
        for cat_name, bindings in data['modes'][''].items():
            for b in bindings:
                if 'alias' in b:
                    alias_entries.append((cat_name, b['alias']))
        # Each alias should appear exactly once
        seen: set[str] = set()
        for cat_name, alias_name in alias_entries:
            self.assertNotIn(alias_name, seen,
                             f'Alias {alias_name!r} appears in multiple places')
            seen.add(alias_name)

    def test_unmapped_actions_sorted_order(self):
        # Covered by test_collect_keys_bindings_sorted
        pass



#!/usr/bin/env python
# License: GPLv3 Copyright: 2021, Kovid Goyal <kovid at kovidgoyal.net>


import json
import os
import shutil
import stat
import subprocess
import sys
import tempfile
import textwrap
import unittest
from functools import partial

from . import BaseTest


class TestBuild(BaseTest):

    def test_exe(self) -> None:
        from kitty.constants import kitten_exe, kitty_exe, str_version
        exe = kitty_exe()
        self.assertTrue(os.access(exe, os.X_OK))
        self.assertTrue(os.path.isfile(exe))
        self.assertIn('kitty', os.path.basename(exe))
        exe = kitten_exe()
        self.assertTrue(os.access(exe, os.X_OK))
        self.assertTrue(os.path.isfile(exe))
        self.assertIn(str_version, subprocess.check_output([exe, '--version']).decode())

    def test_loading_extensions(self) -> None:
        import kitty.fast_data_types as fdt
        from kittens.transfer import rsync
        del fdt, rsync

    def test_loading_shaders(self) -> None:
        from kitty.shaders import Program
        for name in 'cell border bgimage tint graphics'.split():
            Program(name)

    def test_macos_dictation_forwarding(self) -> None:
        from kitty.constants import glfw_path, is_macos
        if not is_macos or not shutil.which('clang'):
                self.skipTest('Dictation smoke test is macOS only and requires clang')
        cocoa_module = glfw_path('cocoa')
        probe = textwrap.dedent('''\
            #import <AppKit/AppKit.h>
            #import <dlfcn.h>
            #import <objc/runtime.h>
            #import <objc/message.h>

            static int start_calls = 0;
            static int stop_calls = 0;
            static id last_sender = nil;

            static void fake_start_dictation(id self, SEL _cmd, id sender) {
                (void)self; (void)_cmd;
                start_calls++;
                last_sender = sender;
            }

            static void fake_stop_dictation(id self, SEL _cmd, id sender) {
                (void)self; (void)_cmd;
                stop_calls++;
                last_sender = sender;
            }

            static void require_true(BOOL condition, const char *message) {
                if (!condition) {
                    fprintf(stderr, "FAIL: %s\\n", message);
                    exit(1);
                }
            }

            int main(void) {
                @autoreleasepool {
                    [NSApplication sharedApplication];
                    void *handle = dlopen(@@COCOA_MODULE@@, RTLD_NOW | RTLD_GLOBAL);
                    require_true(handle != NULL, dlerror());

                    SEL start = NSSelectorFromString(@"startDictation:");
                    SEL stop = NSSelectorFromString(@"stopDictation:");
                    Method start_method = class_getInstanceMethod([NSApplication class], start);
                    Method stop_method = class_getInstanceMethod([NSApplication class], stop);
                    require_true(start_method != NULL, "NSApplication startDictation: missing");
                    require_true(stop_method != NULL, "NSApplication stopDictation: missing");
                    method_setImplementation(start_method, (IMP)fake_start_dictation);
                    method_setImplementation(stop_method, (IMP)fake_stop_dictation);

                    Class view_cls = NSClassFromString(@"GLFWContentView");
                    Class context_cls = NSClassFromString(@"GLFWTextInputContext");
                    require_true(view_cls != Nil, "GLFWContentView class not loaded");
                    require_true(context_cls != Nil, "GLFWTextInputContext class not loaded");

                    SEL init_with_glfw_window = NSSelectorFromString(@"initWithGlfwWindow:");
                    id view = ((id (*)(id, SEL, void *)) objc_msgSend)([view_cls alloc], init_with_glfw_window, NULL);
                    require_true(view != nil, "GLFWContentView initWithGlfwWindow: failed");
                    require_true([view respondsToSelector:start], "GLFWContentView does not expose startDictation:");
                    require_true([view respondsToSelector:stop], "GLFWContentView does not expose stopDictation:");

            #pragma clang diagnostic push
            #pragma clang diagnostic ignored "-Warc-performSelector-leaks"
                    [view performSelector:start withObject:@"menu sender"];
            #pragma clang diagnostic pop
                    require_true(start_calls == 1, "startDictation: action was not forwarded to NSApplication");
                    require_true([(id)last_sender isEqual:@"menu sender"], "startDictation: forwarded wrong sender");

                    [view doCommandBySelector:start];
                    require_true(start_calls == 2, "doCommandBySelector:startDictation: was swallowed");
                    require_true(last_sender == view, "doCommandBySelector:startDictation: should forward self as sender");

                    id context = [view inputContext];
                    require_true(context != nil, "GLFWContentView inputContext missing");
                    require_true([context isKindOfClass:context_cls], "GLFWContentView inputContext has wrong class");
                    [context doCommandBySelector:stop];
                    require_true(stop_calls == 1, "GLFWTextInputContext did not forward stopDictation:");
                    require_true(last_sender == nil, "GLFWTextInputContext should forward nil sender");

                    printf("dictation forwarding probe passed\\n");
                }
                return 0;
            }
        ''').replace('@@COCOA_MODULE@@', json.dumps(cocoa_module))
        with tempfile.TemporaryDirectory() as tdir:
            src = os.path.join(tdir, 'dictation_probe.m')
            exe = os.path.join(tdir, 'dictation_probe')
            with open(src, 'w') as f:
                f.write(probe)
            cp = subprocess.run(
                ['clang', '-framework', 'AppKit', src, '-o', exe],
                stdout=subprocess.PIPE, stderr=subprocess.STDOUT, text=True
            )
            self.assertEqual(cp.returncode, 0, cp.stdout)
            cp = subprocess.run([exe], stdout=subprocess.PIPE, stderr=subprocess.STDOUT, text=True)
            self.assertEqual(cp.returncode, 0, cp.stdout)
            self.assertIn('dictation forwarding probe passed', cp.stdout)

    def test_glfw_modules(self) -> None:
        from kitty.constants import glfw_path, is_macos
        linux_backends = ['x11']
        if not self.is_ci:
            linux_backends.append('wayland')
        modules = ['cocoa'] if is_macos else linux_backends
        for name in modules:
            path = glfw_path(name)
            self.assertTrue(os.path.isfile(path), f'{path} is not a file')
            self.assertTrue(os.access(path, os.X_OK), f'{path} is not executable')

    def test_all_kitten_names(self) -> None:
        from kittens.runner import all_kitten_names
        names = all_kitten_names()
        self.assertIn('diff', names)
        self.assertIn('hints', names)
        self.assertGreater(len(names), 8)

    def test_filesystem_locations(self) -> None:
        from kitty.constants import fonts_dir, local_docs, logo_png_file, shell_integration_dir, terminfo_dir
        zsh = os.path.join(shell_integration_dir, 'zsh')
        self.assertTrue(os.path.isdir(terminfo_dir), f'Terminfo dir: {terminfo_dir}')
        self.assertTrue(os.path.exists(logo_png_file), f'Logo file: {logo_png_file}')
        self.assertTrue(os.path.exists(zsh), f'Shell integration: {zsh}')
        nsfm = os.path.join(fonts_dir, 'SymbolsNerdFontMono-Regular.ttf')
        self.assertTrue(os.path.exists(nsfm), f'Logo file: {nsfm}')

        def is_executable(x):
            mode = os.stat(x).st_mode
            q = stat.S_IXUSR | stat.S_IXGRP | stat.S_IXOTH
            return mode & q == q

        for x in ('kitty', 'kitten'):
            x = os.path.join(shell_integration_dir, 'ssh', x)
            self.assertTrue(is_executable(x), f'{x} is not executable')
        if getattr(sys, 'frozen', False):
            self.assertTrue(os.path.isdir(local_docs()), f'Local docs: {local_docs()}')

    def test_ca_certificates(self):
        import ssl
        if not getattr(sys, 'frozen', False):
            self.skipTest('CA certificates are only tested on frozen builds')
        c = ssl.create_default_context()
        self.assertGreater(c.cert_store_stats()['x509_ca'], 2)

    def test_docs_url(self):
        from kitty.constants import website_url
        from kitty.utils import docs_url

        def run_tests(p, base, suffix='.html'):
            def t(x, e):
                self.ae(p(x), base + e)
            t('', 'index.html' if suffix == '.html' else '')
            t('conf', f'conf{suffix}')
            t('kittens/ssh#frag', f'kittens/ssh{suffix}#frag')
            t('#ref=confloc', f'conf{suffix}#confloc')
            t('#ref=conf-kitty-fonts', f'conf{suffix}#conf-kitty-fonts')
            t('#ref=conf-kitten-ssh-xxx', f'kittens/ssh{suffix}#conf-kitten-ssh-xxx')
            t('#ref=at_close_tab', f'remote-control{suffix}#at-close-tab')
            t('#ref=at-close-tab', f'remote-control{suffix}#at-close-tab')
            t('#ref=action-copy', f'actions{suffix}#copy')
            t('#ref=doc-/marks', f'marks{suffix}')

        run_tests(partial(docs_url, local_docs_root='/docs'), 'file:///docs/')
        w = website_url()
        run_tests(partial(docs_url, local_docs_root=None), w, '/')
        self.ae(docs_url('#ref=issues-123'), 'https://github.com/kovidgoyal/kitty/issues/123')

    def test_launcher_ensures_stdio(self):
        import subprocess

        from kitty.constants import kitty_exe
        exe = kitty_exe()
        cp = subprocess.run([exe, '+runpy', f'''\
import os, sys
if sys.stdin:
    os.close(sys.stdin.fileno())
if sys.stdout:
    os.close(sys.stdout.fileno())
if sys.stderr:
    os.close(sys.stderr.fileno())
os.execlp({exe!r}, 'kitty', '+runpy', 'import sys; raise SystemExit(1 if sys.stdout is None or sys.stdin is None or sys.stderr is None else 0)')
'''])
        self.assertEqual(cp.returncode, 0)


def main() -> None:
    tests = unittest.defaultTestLoader.loadTestsFromTestCase(TestBuild)
    r = unittest.TextTestRunner(verbosity=4)
    result = r.run(tests)
    if result.errors or result.failures:
        raise SystemExit(1)



#!/usr/bin/env python
# License: GPL v3 Copyright: 2018, Kovid Goyal <kovid at kovidgoyal.net>

from kitty.config import defaults
from kitty.fast_data_types import BOTTOM_EDGE, LEFT_EDGE, RIGHT_EDGE, TOP_EDGE, Region
from kitty.layout.base import layout_dimension, lgd
from kitty.layout.interface import Grid, Horizontal, Splits, Stack, Tall
from kitty.layout.splits import Pair
from kitty.types import WindowGeometry
from kitty.window import EdgeWidths
from kitty.window_list import WindowList, reset_group_id_counter

from . import BaseTest


class Window:

    def __init__(self, win_id, overlay_for=None, overlay_window_id=None):
        self.id = win_id
        self.overlay_for = overlay_for
        self.overlay_window_id = overlay_window_id
        self.is_visible_in_layout = True
        self.geometry = WindowGeometry(0, 0, 0, 0, 0, 0)
        self.padding = EdgeWidths()
        self.margin = EdgeWidths()
        self.focused = False

    def focus_changed(self, focused):
        self.focused = focused

    def effective_border(self):
        return 1

    def effective_padding(self, edge):
        return 1

    def effective_margin(self, edge):
        return 1

    def set_visible_in_layout(self, val):
        self.is_visible_in_layout = bool(val)

    def set_geometry(self, geometry):
        self.geometry = geometry


def create_layout(cls, opts=None, border_width=2):
    if opts is None:
        opts = defaults
    ans = cls(1, 1)
    ans.set_active_window_in_os_window = lambda idx: None
    ans.swap_windows_in_os_window = lambda a, b: None
    orig = ans._set_dimensions
    def set_dimensions(all_windows):
        orig(all_windows)
        # we need a non-zero width and height for central
        lgd.central = Region((0, 0, 0, 0, 1, 1))
    ans._set_dimensions = set_dimensions
    return ans


class Tab:

    def active_window_changed(self):
        self.current_layout.update_visibility(self.windows)


def create_windows(layout, num=5):
    t = Tab()
    t.current_layout = layout
    t.windows = ans = WindowList(t)
    ans.tab_mem = t
    reset_group_id_counter()
    for i in range(num):
        ans.add_window(Window(i + 1))
    ans.set_active_group_idx(0)
    return ans


def utils(self, q, windows):
    def ids():
        return [w.id for w in windows.groups]

    def visible_ids():
        return {gr.id for gr in windows.groups if gr.is_visible_in_layout}

    def expect_ids(*a):
        self.assertEqual(tuple(ids()), a)

    def check_visible():
        if q.only_active_window_visible:
            self.ae(visible_ids(), {windows.active_group.id})
        else:
            self.ae(visible_ids(), {gr.id for gr in windows.groups})
    return ids, visible_ids, expect_ids, check_visible


class TestLayout(BaseTest):

    def setUp(self):
        super().setUp()
        self.set_options({'tab_bar_style': 'hidden'})

    def do_ops_test(self, q):
        windows = create_windows(q)
        ids, visible_ids, expect_ids, check_visible = utils(self, q, windows)
        # Test layout
        q(windows)
        self.ae(windows.active_group_idx, 0)
        expect_ids(*range(1, len(windows)+1))
        check_visible()

        # Test nth_window
        for i in range(windows.num_groups):
            q.activate_nth_window(windows, i)
            self.ae(windows.active_group_idx, i)
            expect_ids(*range(1, len(windows)+1))
            check_visible()

        # Test next_window
        for i in range(2 * windows.num_groups):
            expected = (windows.active_group_idx + 1) % windows.num_groups
            q.next_window(windows)
            self.ae(windows.active_group_idx, expected)
            expect_ids(*range(1, len(windows)+1))
            check_visible()

        # Test move_window
        windows.set_active_group_idx(0)
        expect_ids(1, 2, 3, 4, 5)
        q.move_window(windows, 3)
        self.ae(windows.active_group_idx, 3)
        expect_ids(4, 2, 3, 1, 5)
        check_visible()
        windows.set_active_group_idx(0)
        q.move_window(windows, 3)
        expect_ids(*range(1, len(windows)+1))
        check_visible()

        # Test add_window
        windows.set_active_group_idx(4)
        q.add_window(windows, Window(6))
        self.ae(windows.num_groups, 6)
        self.ae(windows.active_group_idx, 5)
        expect_ids(*range(1, windows.num_groups+1))
        check_visible()

        # Test remove_window
        prev_window = windows.active_window
        windows.set_active_group_idx(3)
        self.ae(windows.active_group_idx, 3)
        windows.remove_window(windows.active_window)
        self.ae(windows.active_window, prev_window)
        check_visible()
        expect_ids(1, 2, 3, 5, 6)

        windows.set_active_group_idx(0)
        to_remove = windows.active_window
        windows.set_active_group_idx(3)
        windows.remove_window(to_remove)
        self.ae(windows.active_group_idx, 3)
        check_visible()
        expect_ids(2, 3, 5, 6)

        # Test set_active_window
        for i in range(windows.num_groups):
            windows.set_active_group_idx(i)
            self.ae(i, windows.active_group_idx)
            check_visible()

        # Test

    def do_overlay_test(self, q):
        windows = create_windows(q)
        ids, visible_ids, expect_ids, check_visible = utils(self, q, windows)

        # Test add_window
        w = Window(len(windows) + 1)
        before = windows.active_group_idx
        overlaid_group = before
        overlay_window_id = w.id
        windows.add_window(w, group_of=windows.active_window)
        self.ae(before, windows.active_group_idx)
        self.ae(w, windows.active_window)
        expect_ids(1, 2, 3, 4, 5)
        check_visible()

        # Test layout
        q(windows)
        expect_ids(1, 2, 3, 4, 5)
        check_visible()
        w = Window(len(windows) + 1)
        windows.add_window(w)
        expect_ids(1, 2, 3, 4, 5, 6)
        self.ae(windows.active_group_idx, windows.num_groups - 1)

        # Test nth_window
        for i in range(windows.num_groups):
            q.activate_nth_window(windows, i)
            self.ae(windows.active_group_idx, i)
            if i == overlaid_group:
                self.ae(windows.active_window.id, overlay_window_id)
            expect_ids(1, 2, 3, 4, 5, 6)
            check_visible()

        # Test next_window
        for i in range(windows.num_groups):
            expected = (windows.active_group_idx + 1) % windows.num_groups
            q.next_window(windows)
            self.ae(windows.active_group_idx, expected)
            expect_ids(1, 2, 3, 4, 5, 6)
            check_visible()

        # Test move_window
        windows.set_active_group_idx(overlaid_group)
        expect_ids(1, 2, 3, 4, 5, 6)
        q.move_window(windows, 3)
        self.ae(windows.active_group_idx, 3)
        self.ae(windows.active_window.id, overlay_window_id)
        expect_ids(4, 2, 3, 1, 5, 6)
        check_visible()
        windows.set_active_group_idx(0)
        q.move_window(windows, 3)
        expect_ids(1, 2, 3, 4, 5, 6)
        check_visible()

        # Test set_active_window
        for i in range(windows.num_groups):
            windows.set_active_group_idx(i)
            self.ae(i, windows.active_group_idx)
            if i == overlaid_group:
                self.ae(windows.active_window.id, overlay_window_id)
            check_visible()

        # Test remove_window
        expect_ids(1, 2, 3, 4, 5, 6)
        windows.set_active_group_idx(overlaid_group)
        windows.remove_window(overlay_window_id)
        self.ae(windows.active_group_idx, overlaid_group)
        self.ae(windows.active_window.id, 1)
        expect_ids(1, 2, 3, 4, 5, 6)
        check_visible()

    def test_layout_operations(self):
        for layout_class in (Stack, Horizontal, Tall, Grid):
            q = create_layout(layout_class)
            self.do_ops_test(q)

    def test_overlay_layout_operations(self):
        for layout_class in (Stack, Horizontal, Tall, Grid):
            q = create_layout(layout_class)
            self.do_overlay_test(q)

    def test_splits(self):
        q = create_layout(Splits)
        all_windows = create_windows(q, num=0)
        q.add_window(all_windows, Window(1))
        self.ae(all_windows.active_group_idx, 0)
        q.add_window(all_windows, Window(2), location='vsplit')
        self.ae(all_windows.active_group_idx, 1)
        q(all_windows)
        self.ae(q.pairs_root.pair_for_window(2).horizontal, True)
        q.add_window(all_windows, Window(3), location='hsplit')
        self.ae(q.pairs_root.pair_for_window(2).horizontal, False)
        q.add_window(all_windows, Window(4), location='vsplit')
        windows = list(all_windows)
        windows[0].set_geometry(WindowGeometry(0, 0, 10, 20, 0, 0))
        windows[1].set_geometry(WindowGeometry(11, 0, 20, 10, 0, 0))
        windows[2].set_geometry(WindowGeometry(11, 11, 15, 20, 0, 0))
        windows[3].set_geometry(WindowGeometry(16, 11, 20, 20, 0, 0))
        self.ae(q.neighbors_for_window(windows[0], all_windows), {'right': [2, 3]})
        self.ae(q.neighbors_for_window(windows[1], all_windows), {'left': [1], 'bottom': [3, 4]})
        self.ae(q.neighbors_for_window(windows[2], all_windows), {'left': [1], 'right': [4], 'top': [2]})
        self.ae(q.neighbors_for_window(windows[3], all_windows), {'left': [3], 'top': [2]})

    def test_splits_maximize(self):
        q = create_layout(Splits)
        all_windows = create_windows(q, num=0)
        w1 = Window(1)
        q.add_window(all_windows, w1)
        w2 = Window(2)
        q.add_window(all_windows, w2, location='vsplit')
        w3 = Window(3)
        q.add_window(all_windows, w3, location='hsplit')
        # Layout: w1 | (w2 above w3) — horizontal split at root, vertical split on right
        root = q.pairs_root
        # root is horizontal, containing w1 and [w2/w3 vertical pair]
        self.ae(root.horizontal, True)
        right_pair = root.two if isinstance(root.two, Pair) else root.one
        self.assertIsInstance(right_pair, Pair)

        # Focus window 3 (bottom-right)
        all_windows.set_active_group_idx(all_windows.groups.index(all_windows.group_for_window(w3)))

        # Save original biases
        root_bias_before = root.bias
        right_pair_bias_before = right_pair.bias

        # maximize vertical (fill full height) — affects vertical (horizontal==False) pairs
        result = q.layout_action('maximize', ('vertical',), all_windows)
        self.assertTrue(result)
        # right_pair is vertical (horizontal==False) so its bias should be 0.0 (w3 is in 'two')
        self.ae(right_pair.bias, 0.0)
        # root is horizontal so its bias should be unchanged
        self.ae(root.bias, root_bias_before)
        # _maximized_biases should track w3's vertical maximize
        self.assertIn((all_windows.active_group.id, False), q._maximized_biases)

        # Toggle back
        result = q.layout_action('maximize', ('vertical',), all_windows)
        self.assertTrue(result)
        self.ae(right_pair.bias, right_pair_bias_before)
        self.ae(getattr(q, '_maximized_biases', {}), {})

        # maximize horizontal (fill full width) — affects horizontal pairs
        result = q.layout_action('maximize', ('horizontal',), all_windows)
        self.assertTrue(result)
        # root is horizontal, w3 is under root.two (right side), so bias should be 0.0
        self.ae(root.bias, 0.0)
        # right_pair is vertical, so unchanged
        self.ae(right_pair.bias, right_pair_bias_before)

        # Toggle back
        result = q.layout_action('maximize', ('horizontal',), all_windows)
        self.assertTrue(result)
        self.ae(root.bias, root_bias_before)

    def test_layout_dimension_no_negative_cells(self):
        # Regression test for issue #9946: when window padding exceeds the
        # available space (e.g. after maximize sets a window to minimum width),
        # layout_dimension must not produce a negative cells_per_window value
        # which would cause right < left in the resulting window geometry.
        for length, cell_length, decs in (
            (8, 8, [(5, 5)]),   # padding (10) > length (8) > cell_length (8)
            (6, 8, [(4, 4)]),   # length < cell_length
            (0, 8, [(4, 4)]),   # zero length
            (4, 8, [(3, 3)]),   # space_needed == length, no room for cells
        ):
            result = next(layout_dimension(0, length, cell_length, decs))
            self.assertGreaterEqual(result.cells_per_window, 0,
                f'cells_per_window={result.cells_per_window} < 0 for length={length}, '
                f'cell_length={cell_length}, decs={decs}')
            self.assertGreaterEqual(result.content_size, 0,
                f'content_size={result.content_size} < 0 for length={length}, '
                f'cell_length={cell_length}, decs={decs}')
            # content_pos must be within [0, length]: right edge = content_pos + content_size <= length
            self.assertGreaterEqual(result.content_pos, 0)
            self.assertLessEqual(result.content_pos + result.content_size, length,
                f'right ({result.content_pos + result.content_size}) > length ({length}) for '
                f'cell_length={cell_length}, decs={decs}')

    def test_drag_resize_target_windows(self):
        # Helper: call drag_resize_target_windows with given window and edge flags.
        def drtw(q, all_windows, window, edges):
            return q.drag_resize_target_windows(window, 0, 0, edges, all_windows)

        # --- 2-window horizontal split: A | B ---
        q = create_layout(Splits)
        all_windows = create_windows(q, num=0)
        wA = Window(1)
        q.add_window(all_windows, wA)
        wB = Window(2)
        q.add_window(all_windows, wB, location='vsplit')
        q(all_windows)
        root = q.pairs_root
        self.ae(root.horizontal, True)
        root_id = id(root)

        # Right edge of A (left of divider): divider belongs to root, A is on one-side
        d = drtw(q, all_windows, wA, RIGHT_EDGE)
        self.ae(d.horizontal_id, root_id)
        self.ae(d.width_increases_rightwards, True)

        # Left edge of B (right of divider): same divider, same direction
        d = drtw(q, all_windows, wB, LEFT_EDGE)
        self.ae(d.horizontal_id, root_id)
        self.ae(d.width_increases_rightwards, True)

        # Right edge of B (outer border): direction reversed
        d = drtw(q, all_windows, wB, RIGHT_EDGE)
        self.ae(d.horizontal_id, root_id)
        self.ae(d.width_increases_rightwards, False)

        # --- 2-window vertical split: A / B ---
        q = create_layout(Splits)
        all_windows = create_windows(q, num=0)
        wA = Window(1)
        q.add_window(all_windows, wA)
        wB = Window(2)
        q.add_window(all_windows, wB, location='hsplit')
        q(all_windows)
        root = q.pairs_root
        self.ae(root.horizontal, False)
        root_id = id(root)

        # Bottom edge of A: divider belongs to root, A is in one-side
        d = drtw(q, all_windows, wA, BOTTOM_EDGE)
        self.ae(d.vertical_id, root_id)
        self.ae(d.height_increases_downwards, True)

        # Top edge of B: same divider
        d = drtw(q, all_windows, wB, TOP_EDGE)
        self.ae(d.vertical_id, root_id)
        self.ae(d.height_increases_downwards, True)

        # Bottom edge of B (outer border): direction reversed
        d = drtw(q, all_windows, wB, BOTTOM_EDGE)
        self.ae(d.vertical_id, root_id)
        self.ae(d.height_increases_downwards, False)

        # --- 3-window layout: top_pair(A | B) / C ---
        # root(vertical) -> one: top_pair(horizontal, one=A, two=B), two: C
        q = create_layout(Splits)
        all_windows = create_windows(q, num=0)
        wA = Window(1)
        q.add_window(all_windows, wA)
        wC = Window(3)
        q.add_window(all_windows, wC, location='hsplit')  # C below A (vertical root)
        all_windows.set_active_window_group_for(wA)
        wB = Window(2)
        q.add_window(all_windows, wB, location='vsplit')  # B right of A (horizontal top_pair)
        q(all_windows)
        root = q.pairs_root
        self.ae(root.horizontal, False)
        top_pair = root.one
        self.assertIsInstance(top_pair, Pair)
        self.ae(top_pair.horizontal, True)
        root_id = id(root)
        top_pair_id = id(top_pair)

        # Divider between A and B: belongs to top_pair
        d = drtw(q, all_windows, wA, RIGHT_EDGE)
        self.ae(d.horizontal_id, top_pair_id)
        self.ae(d.width_increases_rightwards, True)

        d = drtw(q, all_windows, wB, LEFT_EDGE)
        self.ae(d.horizontal_id, top_pair_id)
        self.ae(d.width_increases_rightwards, True)

        # Divider between top_pair and C: belongs to root
        d = drtw(q, all_windows, wA, BOTTOM_EDGE)
        self.ae(d.vertical_id, root_id)
        self.ae(d.height_increases_downwards, True)

        d = drtw(q, all_windows, wB, BOTTOM_EDGE)
        self.ae(d.vertical_id, root_id)
        self.ae(d.height_increases_downwards, True)

        d = drtw(q, all_windows, wC, TOP_EDGE)
        self.ae(d.vertical_id, root_id)
        self.ae(d.height_increases_downwards, True)

        # --- 3-window layout: A | right_pair(B / C) ---
        # root(horizontal) -> one: A, two: right_pair(vertical, one=B, two=C)
        q = create_layout(Splits)
        all_windows = create_windows(q, num=0)
        wA = Window(1)
        q.add_window(all_windows, wA)
        wB = Window(2)
        q.add_window(all_windows, wB, location='vsplit')  # B right of A (horizontal root)
        wC = Window(3)
        q.add_window(all_windows, wC, location='hsplit')  # C below B (vertical right_pair)
        q(all_windows)
        root = q.pairs_root
        self.ae(root.horizontal, True)
        right_pair = root.two
        self.assertIsInstance(right_pair, Pair)
        self.ae(right_pair.horizontal, False)
        root_id = id(root)
        right_pair_id = id(right_pair)

        # Divider between A and right_pair: A at RIGHT_EDGE -> root
        d = drtw(q, all_windows, wA, RIGHT_EDGE)
        self.ae(d.horizontal_id, root_id)
        self.ae(d.width_increases_rightwards, True)

        # B at LEFT_EDGE: B is on the leading side of right_pair, border belongs to root
        d = drtw(q, all_windows, wB, LEFT_EDGE)
        self.ae(d.horizontal_id, root_id)
        self.ae(d.width_increases_rightwards, True)

        # Divider between B and C: belongs to right_pair
        d = drtw(q, all_windows, wB, BOTTOM_EDGE)
        self.ae(d.vertical_id, right_pair_id)
        self.ae(d.height_increases_downwards, True)

        d = drtw(q, all_windows, wC, TOP_EDGE)
        self.ae(d.vertical_id, right_pair_id)
        self.ae(d.height_increases_downwards, True)

        # --- 4-window layout (bug scenario): left_pair(A/C) | right_pair(B/D) ---
        # root(horizontal) -> one: left_pair(vertical, one=A, two=C),
        #                      two: right_pair(vertical, one=B, two=D)
        q = create_layout(Splits)
        all_windows = create_windows(q, num=0)
        wA = Window(1)
        q.add_window(all_windows, wA)
        wB = Window(2)
        q.add_window(all_windows, wB, location='vsplit')   # B right of A
        all_windows.set_active_window_group_for(wA)
        wC = Window(3)
        q.add_window(all_windows, wC, location='hsplit')   # C below A
        all_windows.set_active_window_group_for(wB)
        wD = Window(4)
        q.add_window(all_windows, wD, location='hsplit')   # D below B
        q(all_windows)
        root = q.pairs_root
        self.ae(root.horizontal, True)
        left_pair = root.one
        right_pair = root.two
        self.assertIsInstance(left_pair, Pair)
        self.assertIsInstance(right_pair, Pair)
        self.ae(left_pair.horizontal, False)
        self.ae(right_pair.horizontal, False)
        root_id = id(root)
        left_pair_id = id(left_pair)
        right_pair_id = id(right_pair)

        # Bug #1: A at RIGHT_EDGE should give root with correct (rightward) direction
        d = drtw(q, all_windows, wA, RIGHT_EDGE)
        self.ae(d.horizontal_id, root_id)
        self.ae(d.width_increases_rightwards, True)

        # Bug #2: B at LEFT_EDGE should find root and give correct direction
        d = drtw(q, all_windows, wB, LEFT_EDGE)
        self.ae(d.horizontal_id, root_id)
        self.ae(d.width_increases_rightwards, True)

        # C at RIGHT_EDGE: same divider between left_pair and right_pair
        d = drtw(q, all_windows, wC, RIGHT_EDGE)
        self.ae(d.horizontal_id, root_id)
        self.ae(d.width_increases_rightwards, True)

        # D at LEFT_EDGE: same divider
        d = drtw(q, all_windows, wD, LEFT_EDGE)
        self.ae(d.horizontal_id, root_id)
        self.ae(d.width_increases_rightwards, True)

        # Vertical divider within left_pair (between A and C)
        d = drtw(q, all_windows, wA, BOTTOM_EDGE)
        self.ae(d.vertical_id, left_pair_id)
        self.ae(d.height_increases_downwards, True)

        d = drtw(q, all_windows, wC, TOP_EDGE)
        self.ae(d.vertical_id, left_pair_id)
        self.ae(d.height_increases_downwards, True)

        # Vertical divider within right_pair (between B and D)
        d = drtw(q, all_windows, wB, BOTTOM_EDGE)
        self.ae(d.vertical_id, right_pair_id)
        self.ae(d.height_increases_downwards, True)

        d = drtw(q, all_windows, wD, TOP_EDGE)
        self.ae(d.vertical_id, right_pair_id)
        self.ae(d.height_increases_downwards, True)



#!/usr/bin/env python
# vim:fileencoding=utf-8
# License: GPL v3 Copyright: 2017, Kovid Goyal <kovid at kovidgoyal.net>

import json
import os
import shutil
import subprocess
import sys
from copy import deepcopy
from typing import Any

SRC = 'kitty.svg'
base = os.path.dirname(os.path.abspath(__file__))

# To generate this template create an icon using Icon Composer on macOS and in
# the saved .icon (which is a folder) look for icon.js
icon_settings:dict[str, Any] = {
  'fill-specializations' : [
    {
      'value' : {
        'automatic-gradient' : 'extended-gray:1.00000,1.00000'
      }
    },
    {
      'appearance' : 'dark',
      'value' : {
        'automatic-gradient' : 'display-p3:0.20500,0.20500,0.20500,1.00000'
      }
    }
  ],
  'groups' : [
    {
      'layers' : [
        {
          'blend-mode' : 'normal',
          'glass' : False,
          'hidden' : False,
          'image-name' : 'icon.svg',
          'name' : 'icon',
          'opacity' : 1,
          'position' : {
            'scale' : 0.9,
            'translation-in-points' : [
              0,
              0
            ]
          }
        }
      ],
      'shadow' : {
        'kind' : 'neutral',
        'opacity' : 0.5
      },
      'translucency' : {
        'enabled' : True,
        'value' : 0.5
      }
    }
  ],
  'supported-platforms' : {
    'circles' : [
      'watchOS'
    ],
    'squares' : 'shared'
  }
}

def abspath(x: str) -> str:
    return os.path.abspath(os.path.join(base, x))


def run(*args: str) -> None:
    try:
        subprocess.check_call(args)
    except OSError:
        raise SystemExit(f'You are missing the {args[0]} program needed to generate the kitty logo')


def get_svg_viewbox(file_path: str) -> tuple[float, ...]:
    import xml.etree.ElementTree as ET
    tree = ET.parse(file_path)
    root = tree.getroot()
    viewbox = root.get('viewBox')
    if viewbox:
        return tuple(float(x) for x in viewbox.split())
    width = root.get('width')
    height = root.get('height')
    return (0.0, 0.0, float(width or 0), float(height or 0))


def create_icon(name: str, svg_path: str, output_path: str) -> str:
    view_box = get_svg_viewbox(svg_path)
    sz = view_box[-1]
    scale = 0.9 * 1024 / sz
    icon_dir = os.path.join(output_path, f'{name}.icon')
    if os.path.exists(icon_dir):
        shutil.rmtree(icon_dir)
    os.mkdir(icon_dir)
    s = deepcopy(icon_settings)
    for group in s['groups']:
        for layer in group['layers']:
            layer['image-name'] = os.path.basename(svg_path)
            layer['name'] = name
            layer['position']['scale'] = scale
    with open(os.path.join(icon_dir, 'icon.json'), 'w') as f:
        json.dump(s, f, indent=2)
    assets_dir = os.path.join(icon_dir, 'Assets')
    os.mkdir(assets_dir)
    shutil.copy(svg_path, assets_dir)
    return icon_dir


def create_assets() -> None:
    actool = [
        'xcrun', 'actool', '--warnings', '--platform', 'macosx', '--compile', base,
        '--minimum-deployment-target', '15.0', '--output-partial-info-plist', '/dev/stdout',
    ]
    icon = create_icon('kitty', abspath(SRC), base)
    run(*(actool + ['--app-icon', 'kitty', icon]))
    shutil.rmtree(icon)


def render(output: str, sz: int = 256) -> None:
    src = abspath(SRC)
    print(f'Rendering {os.path.basename(src)} at {sz}x{sz}...')
    run('rsvg-convert', '-w', str(sz), '-h', str(sz), '-o', output, src)
    run('optipng', '-quiet', '-o7', '-strip', 'all', output)


def main() -> None:
    if 'darwin' in sys.platform.lower():
        create_assets()
        if sys.argv[-1] == 'remote-macos':
            return
    else:
        run('ssh', 'ox', 'zsh', '-ilc', '~/bin/update-kitty && python3 ~/kitty-src/logo/make.py remote-macos')
        run('rsync', '-avz', '--include=*.icns', '--include=*.car', '--exclude=*', 'ox:~/kitty-src/logo/', base + '/')
    render(abspath('kitty.png'))
    render(abspath('kitty-128.png'), sz=128)


if __name__ == '__main__':
    main()
